/* AL_test: a simple surround OpenAL demo 
   Copyright (C) 2008 Angelo "Encelo" Theodorou
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <SDL/SDL.h>
#include <GL/gl.h>

#include <AL/al.h>
#include <AL/alc.h>
#include <AL/alut.h>

void DrawListener(void);
void DrawSource(ALuint s);

GLfloat dist; /* Source is placed at the corners of a dist X dist square */

int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	int quit = 0;
	int mouse = 0;
	
	const SDL_VideoInfo* info = NULL;
	int width = 640;
	int height = 480;
	int bpp = 0;
	int flags = 0;

	ALCcontext *Context;
	ALCdevice *Device;
	ALuint Buffer;
	ALuint Source;

	int loop = 0;
	int velocity = 0;
	ALfloat pitch = 1.0f;
	ALfloat SourcePos[] = {0.0f, 0.0f, 0.0f};
	ALfloat SourceVel[] = {0.0f, 0.0f, 0.0f};
	ALfloat ListenerPos[] = {0.0f, 0.0f, 0.0f};
	ALfloat ListenerVel[] = {0.0f, 0.0f, 0.0f};
	ALfloat ListenerOri[] = {0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f}; /* Y-axis determines front/rear channels */


	if (argc > 2) {
		fprintf(stderr, "Usage: %s distance \n", argv[0]);
		exit(-1);
	}
	else if (argc == 2) {
		if(sscanf(argv[1], "%f", &dist) != 1) {
			fprintf(stderr, "Distance parameter should be a float\n");
			exit(-1);
		}
	}
	else
		dist = 1.0f; /* default value */

	
/* ----- SDL init --------------- */
	if(SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);
	
	info = SDL_GetVideoInfo();
	bpp = info->vfmt->BitsPerPixel;

	
/* ----- OpenGL attribute setting via SDL --------------- */
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	
/*	flags = SDL_OPENGL | SDL_FULLSCREEN; */
	flags = SDL_OPENGL;

	
/* ----- Setting up the screen surface --------------- */
	if((screen = SDL_SetVideoMode(width, height, bpp, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("AL_test by Encelo", NULL);


/* ----- OpenGL init --------------- */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.1f*dist, 1.1f*dist, -1.1f*dist, 1.1f*dist, 0.0f, 100.0f);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);


/* ----- OpenAL init --------------- */
	alutInit (&argc, argv);
	alGetError();

	alGenBuffers(1, &Buffer);
	Buffer = alutCreateBufferFromFile("test.wav");
	if (Buffer == AL_NONE) {
		fprintf(stderr, "Could not load %s\n", "test.wav");
		exit(-1);
	}

	alGenSources(1, &Source);
	alSourcei (Source, AL_BUFFER, Buffer);
	alSourcef (Source, AL_PITCH, pitch);
	alSourcef (Source, AL_GAIN, 1.0);
	alSourcefv(Source, AL_POSITION, SourcePos);
	alSourcefv(Source, AL_VELOCITY, SourceVel);
	alSourcei (Source, AL_LOOPING, AL_FALSE);

	alListenerfv(AL_POSITION, ListenerPos);
	alListenerfv(AL_VELOCITY, ListenerVel);
	alListenerfv(AL_ORIENTATION, ListenerOri);

	Context = alcGetCurrentContext();
	Device = alcGetContextsDevice(Context);
	printf("Default Device Specifier: %s\n", alcGetString(AL_NONE, ALC_DEFAULT_DEVICE_SPECIFIER));
	printf("Default Capture Device Specifier: %s\n", alcGetString(AL_NONE, ALC_CAPTURE_DEFAULT_DEVICE_SPECIFIER));
	printf("Current Device Specifier: %s\n", alcGetString(Device, ALC_DEVICE_SPECIFIER));
	printf("Current Capture Device Specifier: %s\n", alcGetString(Device, ALC_CAPTURE_DEVICE_SPECIFIER));
	printf("ALC Extensions: %s\n", alcGetString(Device, ALC_EXTENSIONS));


/* ----- Event cycle --------------- */
	while (!quit) {
		SourceVel[0] = 0.0f; SourceVel[1] = 0.0f;
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_QUIT:
				quit = 1;
				break;
			case SDL_MOUSEMOTION:
				if (!mouse)
					continue;
				/* event.motion.? / screen dimension translate pixel coordinates into the range of [0..1]
				 * the subtraction that come next further translate it to the [-0.5 .. 0.5] range.
				 * This value is then multiplicated by 2.2 times the distance, to match the OpenGL boundary. */
				SourcePos[0] = (event.motion.x/(float)width - 0.5) * 2.2f*dist; SourcePos[1] = (-event.motion.y/(float)height + 0.5) * 2.2f*dist;
				alSourcefv(Source, AL_POSITION, SourcePos);
				if (velocity)
					/* The velocity is calculated as for the position, but this time using event.motion.?rel, that is the relative movement.
					 * The result is divided by 50, because the difference in the position is polled every 50th of second ( SDL_Delay(20) ). */
					SourceVel[0] = (event.motion.xrel/(float)width) * 2.2f*dist / 50; SourceVel[1] = (-event.motion.yrel/(float)height) * 2.2f*dist / 50;
				break;
			case SDL_KEYDOWN:
				switch(event.key.keysym.sym){
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = 1;
					break;
				case SDLK_F1:
					SDL_WM_ToggleFullScreen(screen);
					break;
				case SDLK_SPACE:
					alSourcePlay(Source);
					break;
				case SDLK_m:
					mouse = !mouse;
					break;
				case SDLK_l:
					loop = !loop;
					if (loop)
						alSourcei (Source, AL_LOOPING, AL_TRUE);
					else
						alSourcei (Source, AL_LOOPING, AL_FALSE);
					break;
				case SDLK_v:
					velocity = !velocity;
					break;
				case SDLK_RIGHT:
					pitch += 0.1f;
					if (pitch < 0.0f)
						pitch = 0.0f;
					break;
				case SDLK_DOWN:
					pitch = 1.0f;
				case SDLK_LEFT:
					pitch -= 0.1f;
					break;

				case SDLK_KP7:
					SourcePos[0] = -dist; SourcePos[1] = dist;
					break;
				case SDLK_KP8:
					SourcePos[0] = 0.0f; SourcePos[1] = dist;
					break;
				case SDLK_KP9:
					SourcePos[0] = dist; SourcePos[1] = dist;
					break;
				case SDLK_KP4:
					SourcePos[0] = -dist; SourcePos[1] = 0.0f;
					break;
				case SDLK_KP5:
					SourcePos[0] = 0.0f; SourcePos[1] = 0.0f;
					break;
				case SDLK_KP6:
					SourcePos[0] = dist; SourcePos[1] = 0.0f;
					break;
				case SDLK_KP1:
					SourcePos[0] = -dist; SourcePos[1] = -dist;
					break;
				case SDLK_KP2:
					SourcePos[0] = 0.0f; SourcePos[1] = -dist;
					break;
				case SDLK_KP3:
					SourcePos[0] = dist; SourcePos[1] = -dist;
					break;
				}
			alSourcefv(Source, AL_POSITION, SourcePos);
			alSourcef (Source, AL_PITCH, pitch);
			}
		alSourcefv(Source, AL_VELOCITY, SourceVel);
		}


/* ----- Blitting on the screen --------------- */
		glClear(GL_COLOR_BUFFER_BIT);
		glLoadIdentity();

		DrawListener();
		DrawSource(Source);

		SDL_GL_SwapBuffers();		
		SDL_Delay(20); /* Decomment this if you want 1/50th screen update */
	}

	
/* ----- Quitting --------------- */
	alDeleteSources(1, &Source);
	alDeleteBuffers(1, &Buffer);
	alutExit();
	SDL_Quit();
	return 0;	
}


void DrawListener(void)
{
	ALfloat x, y, z;


	alGetListener3f(AL_POSITION, &x, &y, &z);

	glPushAttrib(GL_COLOR_BUFFER_BIT);
	glColor3f(0.0f, 0.0f, 1.0f);

	glPushMatrix();
	glTranslatef(x, y, z);
	glScalef(0.1f*dist, 0.1f*dist, 0.1f*dist);
	glBegin(GL_QUADS);
	glVertex3f( 1.0f, 1.0f, 0.0f);
	glVertex3f(-1.0f, 1.0f, 0.0f);
	glVertex3f(-1.0f,-1.0f, 0.0f);
	glVertex3f( 1.0f,-1.0f, 0.0f);
	glEnd();
	glPopMatrix();

	glPopAttrib();
}


void DrawSource(ALuint s)
{
	ALfloat  x, y, z;


	alGetSource3f(s, AL_POSITION, &x, &y, &z);

	glPushAttrib(GL_COLOR_BUFFER_BIT);
	glColor3f(1.0f, 0.0f, 0.0f);

	glPushMatrix();
	glTranslatef(x, y, z);
	glScalef(0.075f*dist, 0.075f*dist, 0.075f*dist);
	glBegin(GL_TRIANGLES);
	glVertex3f( 0.0f, 1.0f, 0.0f);
	glVertex3f(-1.0f,-1.0f, 0.0f);
	glVertex3f( 1.0f, -1.0f, 0.0f);
	glEnd();
	glPopMatrix();

	glPopAttrib();
}
